import os
import pathlib
import pandas as pd
from surprise import Dataset
from surprise import Reader

def load_test_dataset():
    """
    Load a sample of the ml_100k dataset
    """
    file_path = os.path.expanduser('./tests/u1_ml_100k_test')
    reader = Reader(line_format='user item rating timestamp', sep=',', rating_scale=(1, 5))

    return Dataset.load_from_file(file_path, reader=reader)

def get_tmp_dir():
    """
    Create and return tmp dir
    """

    tmp_dir = pathlib.Path().absolute() / "tmp"
    tmp_dir.mkdir(parents=True, exist_ok=True)
    return tmp_dir



class Error(Exception):
   """Base class for other exceptions"""
   pass

class ValidationError(Error):
    """
    Exceptions raised for validation errors
    """

    def __init__(self, expression, message):
        self.expression = expression
        self.message = message



from hyperopt import fmin, tpe, hp, STATUS_OK, Trials
from surprise import KNNWithZScore
from surprise.model_selection import cross_validate
from auto_surprise.constants import DEFAULT_MAX_EVALS, ACCURACY_METRICS
from auto_surprise.algorithms.spaces import KNN_DEFAULT_SPACE
from auto_surprise.algorithms.base import AlgorithmBase


class AutoSurpriseKNNWithZScore(AlgorithmBase):
    """
    Wrapper for surprise.prediction_algorithms.knns.KNNWithZScore
    """

    def _hyperopt(self, params):
        algo = KNNWithZScore(**params)
        return cross_validate(
            algo,
            self._data,
            measures=ACCURACY_METRICS,
            cv=self._cv,
            n_jobs=self._cv_n_jobs,
            verbose=self.verbose,
        )[self._metric].mean()

    def _objective(self, params):
        loss = self._hyperopt(params)
        self._result_logger.append_results(loss, params)

        return {"loss": loss, "status": STATUS_OK, "hyperparams": params}

    def best_hyperparams(self, max_evals):
        best = fmin(
            self._objective,
            KNN_DEFAULT_SPACE,
            algo=self._hpo_algo,
            max_evals=max_evals,
            trials=self.trials,
            verbose=self.verbose,
        )
        return best, self.trials



"""
Spaces defined for the algorithms to user
It might be a good idea to experiment with some of the distributions here
"""

from math import log
from hyperopt import hp
from hyperopt.pyll.base import scope

LR_LOG_LOWER = log(0.0001)
LR_LOG_UPPER = log(0.1)

# NOTE: Do not use keyword arguments for hyperopt distributions. It causes a crash when using ATPE

SVD_DEFAULT_SPACE = {
    "n_factors": scope.int(hp.quniform("n_factors", 1, 200, 1)),
    "n_epochs": scope.int(hp.quniform("n_epochs", 1, 200, 1)),
    "lr_bu": hp.loguniform("lr_bu", LR_LOG_LOWER, LR_LOG_UPPER),
    "lr_bi": hp.loguniform("lr_bi", LR_LOG_LOWER, LR_LOG_UPPER),
    "lr_pu": hp.loguniform("lr_pu", LR_LOG_LOWER, LR_LOG_UPPER),
    "lr_qi": hp.loguniform("lr_qi", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_bu": hp.loguniform("reg_bu", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_bi": hp.loguniform("reg_bi", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_pu": hp.loguniform("reg_pu", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_qi": hp.loguniform("reg_qi", LR_LOG_LOWER, LR_LOG_UPPER),
}

SVDPP_SPACE = {
    **SVD_DEFAULT_SPACE,
    "lr_yj": hp.loguniform("lr_yj", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_yj": hp.loguniform("reg_yj", LR_LOG_LOWER, LR_LOG_UPPER),
}

# TODO: Shrinkage param is only used when pearson_baseline is selected. It would
# be great if we could conditionally optimize if pearson_baseline is used
SIMILARITY_OPTIONS_SPACE = {
    "name": hp.choice("name", ["cosine", "msd", "pearson", "pearson_baseline"]),
    "user_based": hp.choice("user_based", [False, True]),
    "min_support": scope.int(hp.quniform("min_support", 1, 100, 1)),
    "shrinkage": scope.int(hp.quniform("shrinkage", 1, 300, 1)),
}

BSL_OPTIONS_SPACE = hp.choice(
    "bsl_options",
    [
        {
            "method": "als",
            "reg_i": hp.uniform("reg_i", 1, 100),
            "reg_u": hp.uniform("reg_u", 1, 100),
            "n_epochs": scope.int(hp.quniform("n_epochs", 5, 200, 1)),
        },
        {
            "method": "sgd",
            "reg": hp.loguniform("reg", LR_LOG_LOWER, LR_LOG_UPPER),
            "learning_rate": hp.loguniform("learning_rate", LR_LOG_LOWER, LR_LOG_UPPER),
        },
    ],
)

BASELINE_ONLY_SPACE = {"bsl_options": BSL_OPTIONS_SPACE}

KNN_DEFAULT_SPACE = {
    "k": scope.int(hp.quniform("k", 1, 500, 1)),
    "min_k": scope.int(hp.quniform("min_k", 1, 50, 1)),
    "sim_options": SIMILARITY_OPTIONS_SPACE,
}

KNN_BASELINE_SPACE = {**KNN_DEFAULT_SPACE, "bsl_options": BSL_OPTIONS_SPACE}

NMF_DEFAULT_SPACE = {
    "n_factors": scope.int(hp.quniform("n_factors", 1, 500, 1)),
    "n_epochs": scope.int(hp.quniform("n_epochs", 5, 200, 1)),
    "lr_bu": hp.loguniform("lr_bu", LR_LOG_LOWER, LR_LOG_UPPER),
    "lr_bi": hp.loguniform("lr_bi", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_bu": hp.loguniform("reg_bu", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_bi": hp.loguniform("reg_bi", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_pu": hp.loguniform("reg_pu", LR_LOG_LOWER, LR_LOG_UPPER),
    "reg_qi": hp.loguniform("reg_qi", LR_LOG_LOWER, LR_LOG_UPPER),
    "biased": hp.choice("biased", [False, True]),
}

CO_CLUSTERING_DEFAULT_SPACE = {
    "n_cltr_u": scope.int(hp.quniform("n_cltr_u", 1, 1000, 1)),
    "n_cltr_i": scope.int(hp.quniform("n_cltr_i", 1, 100, 1)),
    "n_epochs": scope.int(hp.quniform("n_epochs", 5, 200, 1)),
}



import logging
import multiprocessing
from auto_surprise.constants import EVALS_MULTIPLIER, MAX_WORKERS
from auto_surprise.trainer import Trainer
from auto_surprise.strategies.base import StrategyBase


class ContinuousParallel(StrategyBase):
    """
    Executes all alogrithms in parallel until time limit exceeded or max evals reached.
    """

    def __init__(self, *args, **kwargs):
        self.__logger = logging.getLogger(__name__)
        super().__init__(*args, **kwargs)

    def evaluate(self):
        """
        Evaluate performance of algorithms
        """

        self.__logger.info("Starting evaluation using strategy : ContinuousParallel")

        max_evals = self.max_evals
        processes = []

        # Run for N algorithms
        with multiprocessing.Manager() as mp_manager:

            tasks = mp_manager.dict()

            for algo in self.algorithms:
                if self.verbose:
                    print("Starting process with %s algorithm" % algo)

                trainer = Trainer(
                    self.tmp_dir,
                    algo=algo,
                    data=self.data,
                    target_metric=self.target_metric,
                    hpo_algo=self.hpo_algo,
                    verbose=self.verbose,
                )
                p = multiprocessing.Process(
                    target=trainer.start_with_limits,
                    args=(max_evals, self.time_limit, tasks),
                )
                processes.append(p)
                p.start()

            # Wait for processes to complete
            for process in processes:
                process.join()

            passed_jobs = {k:v for k,v in tasks.items() if v["loss"]}
            best_algo = min(passed_jobs.items(), key=(lambda x: x[1]["loss"]))[0]
            best_params = passed_jobs[best_algo]["hyperparams"]
            best_score = passed_jobs[best_algo]["loss"]

            return best_algo, best_params, best_score, tasks.copy()



import unittest
import auto_surprise.validation_util as validation_util
from auto_surprise.exceptions import ValidationError

import utils as test_utils

class TestValidationUtil(unittest.TestCase):
    def test_validate_target_metric(self):
        """
        Test that a valid target metric is selected
        If the metric is not valid, an exception should be thrown
        """

        valid_metric = "test_rmse"
        self.assertTrue(validation_util.validate_target_metric(valid_metric))

        invalid_metric = "recall"
        self.assertRaises(ValidationError, validation_util.validate_target_metric, invalid_metric)

    def test_validate_dataset(self):
        """
        Validate that the given dataset is in the required format
        """

        self.assertRaises(ValidationError, validation_util.validate_dataset, None)
        self.assertRaises(ValidationError, validation_util.validate_dataset, True)
        valid_dataset = test_utils.load_test_dataset()
        self.assertTrue(validation_util.validate_dataset(valid_dataset))

    def test_validate_max_evals(self):
        """
        Validate that max evals is an integer greater than 0
        """

        self.assertRaises(ValidationError, validation_util.validate_max_evals, "3.14")
        self.assertTrue(validation_util.validate_max_evals(10))

    def test_validate_cpu_time_limit(self):
        """
        Validate that max evals is an integer greater than 0
        """

        self.assertRaises(ValidationError, validation_util.validate_cpu_time_limit, "0.451")
        self.assertTrue(validation_util.validate_cpu_time_limit(10))



import signal
from contextlib import contextmanager

class TimeoutException(Exception): pass

@contextmanager
def run_with_enforced_limits(time_limit=None):
    """
    Execute methods with resource limits. Works only for linux.
    """

    def signal_handler(signum, frame):
        raise TimeoutException("This job has timed out. The results will still be used")

    signal.signal(signal.SIGALRM, signal_handler)

    # Set alarm for time limit
    signal.alarm(time_limit)

    try:
        yield
    finally:
        signal.alarm(0)



from auto_surprise.constants import AVAILABLE_METRICS
from auto_surprise.exceptions import ValidationError
from surprise.dataset import DatasetAutoFolds


def validate_target_metric(metric):
    if metric not in AVAILABLE_METRICS:
        raise ValidationError(
            "target_metric",
            'Unknown target metric : "%s". Must be in %s' % (metric, AVAILABLE_METRICS),
        )
    else:
        return True


def validate_dataset(data):
    """
    Validate user given dataset
    """
    if data:
        if not isinstance(data, DatasetAutoFolds):
            raise ValidationError(
                "data",
                'Unknown data format. Must be and instance of "DatasetAutoFolds". Got "%s"'
                % type(data),
            )
        else:
            return True
    else:
        raise ValidationError("data", "Data must be present. Got: %s" % data)


def validate_max_evals(max_evals):
    """
    Max evaluations for hyperparameter tuning. Must be an unsigned integer greater than 0
    """
    if not isinstance(max_evals, int) or not max_evals > 0:
        raise ValidationError(
            "max_evals",
            "max_evals must be an unsigned integer greater than 0. Got {0}".format(
                max_evals
            ),
        )
    else:
        return True

def validate_cpu_time_limit(cpu_time_limit):
    """
    Execution time limit. Must be an unsigned integer greater than 0
    """
    if not isinstance(cpu_time_limit, int) or not cpu_time_limit > 0:
        raise ValidationError(
            "cpu_time_limit",
            "cpu_time_limit must be an unsigned integer greater than 0. Got {0}".format(
                cpu_time_limit
            ),
        )
    else:
        return True



import unittest
import pathlib
import utils as test_utils
from auto_surprise.trainer import Trainer

class TestTrainer(unittest.TestCase):
    def setUp(self):
        self.data = test_utils.load_test_dataset()
        self.tmp_path = test_utils.get_tmp_dir()

    def test_start_with_limits(self):
        trainer = Trainer(self.tmp_path, algo="svd", data=self.data)

        with self.subTest(msg="With no exceptions, trainer should populate tasks"):
            tasks = {}
            trainer.start_with_limits(2, 10, tasks)
            self.assertTrue(tasks["svd"])
            self.assertListEqual(list(tasks["svd"].keys()), ["loss", "status", "hyperparams", "exception"])

        with self.subTest(msg="With timeout exception, trainer should still populate tasks with exception tag false"):
            tasks = {}
            trainer.start_with_limits(100, 10, tasks)
            self.assertTrue(tasks["svd"])
            self.assertEqual(tasks["svd"]["exception"], False)

        with self.subTest(msg="With timeout exception, but the job was not evaluated even once"):
            tasks = {}
            trainer = Trainer(self.tmp_path, algo="knn_baseline", data=self.data)           
            trainer.start_with_limits(100, 1, tasks)
            self.assertTrue(tasks["knn_baseline"])
            self.assertEqual(tasks["knn_baseline"]["exception"], False)
            # This is a bit hard to assert, since sometimes it can evaluate in 1 sec. Need to revisit this test later
            # self.assertIsNone(tasks["knn_baseline"]["loss"])



from surprise import SVD
from surprise import SVDpp
from surprise import NMF
from surprise import KNNBaseline
from surprise import KNNBasic
from surprise import KNNWithMeans
from surprise import KNNWithZScore
from surprise import SlopeOne
from surprise import CoClustering
from surprise import NormalPredictor
from surprise import BaselineOnly
from hyperopt import tpe
### Algorithm maps

ALGORITHM_MAP = {
    'svd': 'AutoSurpriseSVD',
    'svdpp': 'AutoSurpriseSVDpp',
    'nmf': 'AutoSurpriseNMF',
    'knn_baseline': 'AutoSurpriseKNNBaseline',
    'knn_basic': 'AutoSurpriseKNNBasic',
    'knn_with_means': 'AutoSurpriseKNNWithMeans',
    'knn_with_z_score': 'AutoSurpriseKNNWithZScore',
    'co_clustering': 'AutoSurpriseCoClustering',
    'slope_one': 'AutoSurpriseSlopeOne',
    'baseline_only': 'AutoSurpriseBaselineOnly',
    'normal_predictor': 'AutoSurpriseNormalPredictor',
}

SURPRISE_ALGORITHM_MAP = {
    'svd': SVD,
    'svdpp': SVDpp,
    'nmf': NMF,
    'knn_baseline': KNNBaseline,
    'knn_basic': KNNBasic,
    'knn_with_means': KNNWithMeans,
    'knn_with_z_score': KNNWithZScore,
    'co_clustering': CoClustering,
    'slope_one': SlopeOne,
    'baseline_only': BaselineOnly,
    'normal_predictor': NormalPredictor,
}

FULL_ALGO_LIST = ['svd', 'svdpp', 'nmf', 'knn_basic', 'knn_baseline', 'knn_with_means', 'knn_with_z_score', 'co_clustering', 'slope_one', 'baseline_only']
QUICK_COMPUTE_ALGO_LIST = ['svd', 'nmf', 'knn_basic', 'knn_baseline', 'knn_with_means', 'knn_with_z_score', 'co_clustering', 'slope_one', 'baseline_only']
BASELINE_ALGO = 'normal_predictor'

### Metrics

DEFAULT_TARGET_METRIC = "test_rmse"

AVAILABLE_METRICS = [
    "test_rmse",
    "test_mae",
    "test_mse",
]

ACCURACY_METRICS = [
    "RMSE",
    "MAE",
    "MSE",
]

### Run params
DEFAULT_CV_ITERS = 5
DEFAULT_MAX_EVALS = 10
EVALS_MULTIPLIER = 5
MAX_WORKERS = None
DEFAULT_HPO_ALGO = tpe.suggest
CV_N_JOBS = 1



import os
import logging
import pathlib

from auto_surprise.constants import (
    DEFAULT_TARGET_METRIC,
    DEFAULT_MAX_EVALS,
    FULL_ALGO_LIST,
    BASELINE_ALGO,
    EVALS_MULTIPLIER,
    SURPRISE_ALGORITHM_MAP,
    DEFAULT_HPO_ALGO,
)
from auto_surprise.trainer import Trainer
from auto_surprise.exceptions import ValidationError
from auto_surprise.context.backend import BackendContextManager
from auto_surprise.strategies.continuous_parallel import ContinuousParallel
import auto_surprise.validation_util as validation_util


class Engine(object):
    def __init__(self, verbose=True, algorithms=FULL_ALGO_LIST):
        """
        Initialize new engine
        """
        self.__logger = logging.getLogger(__name__)
        self.verbose = verbose
        self.algorithms = algorithms
        self._current_path = pathlib.Path().absolute()

    def train(
        self,
        target_metric=DEFAULT_TARGET_METRIC,
        data=None,
        max_evals=DEFAULT_MAX_EVALS,
        cpu_time_limit=None,
        hpo_algo=DEFAULT_HPO_ALGO,
    ):
        """
        Train and find most optimal model and hyperparameters
        """

        try:
            # Validations
            validation_util.validate_target_metric(target_metric)
            validation_util.validate_dataset(data)
            validation_util.validate_max_evals(max_evals)
            validation_util.validate_cpu_time_limit(cpu_time_limit)
        except ValidationError as err:
            """
            Catch validation errors. Auto-Surprise cannot run with these exceptions
            """
            self.__logger.critical(err.message)
            raise err

        # Determine baseline value from random normal predictor
        with BackendContextManager(self._current_path) as tmp_dir:
            if self.verbose:
                print("Available CPUs: {0}".format(os.cpu_count()))

            # Calculate baseline first. This can be used to early stop training of algorithms if results are not optimal
            baseline_trainer = Trainer(
                tmp_dir,
                algo=BASELINE_ALGO,
                data=data,
                target_metric=target_metric,
                verbose=self.verbose,
            )
            baseline_loss = baseline_trainer.start(1)[1]["loss"]
            if self.verbose:
                print("Baseline loss : {0}".format(baseline_loss))

            # Initialize the strategy to be used to optimize. Currently only one strategy implemented.
            strategy = ContinuousParallel(
                self.algorithms,
                data,
                target_metric,
                baseline_loss,
                tmp_dir,
                time_limit=cpu_time_limit,
                max_evals=max_evals,
                hpo_algo=hpo_algo,
                verbose=self.verbose,
            )

            best_algo, best_params, best_score, tasks = strategy.evaluate()

        if self.verbose:
            print("----Done!----")
            print("Best algorithm: {0}".format(best_algo))
            print("Best hyperparameters: {0}".format(best_params))

        return best_algo, best_params, best_score, tasks

    def build_model(self, algo_name, params):
        algo = SURPRISE_ALGORITHM_MAP[algo_name]
        if params:
            return algo(**params)
        else:
            return algo()



from auto_surprise.constants import DEFAULT_MAX_EVALS, DEFAULT_HPO_ALGO

class StrategyBase():
    def __init__(
        self,
        algorithms,
        data,
        target_metric,
        baseline_loss,
        temporary_directory,
        time_limit=None,
        max_evals=DEFAULT_MAX_EVALS,
        hpo_algo=DEFAULT_HPO_ALGO,
        verbose=False,
    ):
        self.algorithms = algorithms
        self.data = data
        self.target_metric = target_metric
        self.baseline_loss = baseline_loss
        self.time_limit = time_limit
        self.tmp_dir = temporary_directory
        self.hpo_algo = hpo_algo
        self.max_evals = max_evals
        self.verbose = verbose



from datetime import datetime
import logging
import pathlib


class BackendContextManager:
    def __init__(self, current_path):
        date_time = datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        self.temporary_directory = current_path / "tmp" / date_time
        self.__logger = logging.getLogger(__name__)

    def __enter__(self):
        self.create_directories()
        return self.temporary_directory

    def __exit__(self, *exc):
        self.__logger.info(
            "Stored temporary files in %s. This can be removed"
            % (self.temporary_directory.as_posix())
        )

    def create_directories(self):
        self.temporary_directory.mkdir(parents=True)



import time
import datetime
import pathlib
import logging

class ResultLoggingManager:
    """
    Log results of each algorithm
    """

    def __init__(self, tmp_dir, algo_name):
        self._start_time = time.time()
        self._log_file_path = tmp_dir / "{0}_results.log".format(algo_name)

        self.logger = logging.getLogger("{0}_{1}".format(__name__, algo_name))
        self.logger.setLevel(logging.INFO)

        file_handler = logging.FileHandler(self._log_file_path)
        formatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)

    def __enter__(self):
        self.logger.info("Started hyperarameter optimization")
        return self

    def __exit__(self, *exc):
        self.logger.info("Done!")

    def append_results(self, loss, hyperparams):
        execution_time = str(
            datetime.timedelta(seconds=int(time.time() - self._start_time))
        )
        self.logger.info(
            "Execution time: {0} Loss: {1} Hyperparameters: {2}".format(
                execution_time, loss, hyperparams
            )
        )


from auto_surprise.algorithms.auto_surprise_svd import AutoSurpriseSVD
from auto_surprise.algorithms.auto_surprise_svdpp import AutoSurpriseSVDpp
from auto_surprise.algorithms.auto_surprise_nmf import AutoSurpriseNMF
from auto_surprise.algorithms.auto_surprise_knn_baseline import AutoSurpriseKNNBaseline
from auto_surprise.algorithms.auto_surprise_knn_basic import AutoSurpriseKNNBasic
from auto_surprise.algorithms.auto_surprise_knn_with_means import AutoSurpriseKNNWithMeans
from auto_surprise.algorithms.auto_surprise_knn_with_z_score import AutoSurpriseKNNWithZScore
from auto_surprise.algorithms.auto_surprise_slope_one import AutoSurpriseSlopeOne
from auto_surprise.algorithms.auto_surprise_co_clustering import AutoSurpriseCoClustering
from auto_surprise.algorithms.auto_surprise_normal_predictor import AutoSurpriseNormalPredictor
from auto_surprise.algorithms.auto_surprise_baseline_only import AutoSurpriseBaselineOnly


import concurrent.futures

from auto_surprise.constants import (EVALS_MULTIPLIER, MAX_WORKERS)
from auto_surprise.trainer import Trainer
from auto_surprise.strategies.base import StrategyBase

class BasicReduction(StrategyBase):
    """
    A basic strategy for comparison of algorithms
    """

    def evaluate(self):
        """
        Evaluate performance of algorithms
        """

        print("Starting evaluation using strategy : BasicReduction")

        tasks = {}
        iteration = 0
        while True:
            print("Iteration: %i" % iteration)

            max_evals = EVALS_MULTIPLIER * (iteration + 1)
            tasks[iteration] = {}
            futures = {}
            with concurrent.futures.ProcessPoolExecutor(max_workers=MAX_WORKERS) as executor:
                # Run for N algorithms
                for algo in self.algorithms:
                    print("Starting thread with %s algorithm" % algo)

                    trainer = Trainer(algo=algo, data=self.data, target_metric=self.target_metric, hpo_algo=self.hpo_algo, debug=self._debug)
                    futures[executor.submit(trainer.start, max_evals)] = algo

                # Load results of completed tasks
                for future in concurrent.futures.as_completed(futures):
                    algo = futures[future]
                    hyperparams, score = future.result()

                    # If no exceptions, then include in tasks, else remove from algorithms list
                    if hyperparams or score:
                        tasks[iteration][algo] = {
                            'hyperparameters': hyperparams,
                            'score': score,
                            'above_baseline': score['loss'] < self.baseline_loss
                        }
                    else:
                        print('Cannot use algo : %s' % algo)

                        tasks[iteration][algo] = {
                            'above_baseline': False,
                            'exception': True
                        }

            if len(tasks[iteration]) == 1:
                break
            else:
                self.__filter_algorithms(tasks[iteration])
                iteration += 1

        best_model = list(tasks[iteration].keys())[0]
        best_params = tasks[iteration][best_model]['hyperparameters']
        best_score = tasks[iteration][best_model]['score']['loss']

        return best_model, best_params, best_score, tasks

    def __filter_algorithms(self, tasks):
        """
        Rank N algorithms and take the top N/2 algorithms which performed
        better than baseline result for the next iteration
        """

        filtered_algorithms = dict(filter(lambda algo: algo[1]['above_baseline'], tasks.items()))
        algorithms_ranking = [i[0] for i in sorted(filtered_algorithms.items(), key=lambda x: x[1]['score']['loss'], reverse=False)]
        algorithms_count = round(len(self.algorithms) / 2)
        algorithms = algorithms_ranking[0:algorithms_count]

        return algorithms



import sys
import traceback
import logging

from surprise.model_selection import cross_validate
from auto_surprise.algorithms import (
    AutoSurpriseBaselineOnly,
    AutoSurpriseCoClustering,
    AutoSurpriseKNNBaseline,
    AutoSurpriseKNNBasic,
    AutoSurpriseKNNWithMeans,
    AutoSurpriseKNNWithZScore,
    AutoSurpriseNMF,
    AutoSurpriseSlopeOne,
    AutoSurpriseSVD,
    AutoSurpriseSVDpp,
    AutoSurpriseNormalPredictor,
)
from auto_surprise.constants import (
    ALGORITHM_MAP,
    DEFAULT_TARGET_METRIC,
    CV_N_JOBS,
    DEFAULT_HPO_ALGO,
)
from auto_surprise.context.limits import TimeoutException
from auto_surprise.context.result_logging_manager import ResultLoggingManager
import auto_surprise.context.limits as limits


class Trainer(object):
    def __init__(
        self,
        tmp_dir,
        algo=None,
        data=None,
        target_metric=DEFAULT_TARGET_METRIC,
        hpo_algo=DEFAULT_HPO_ALGO,
        verbose=False,
    ):
        """
        Initialize new trainer
        """
        self.__logger = logging.getLogger(__name__)
        self.verbose = verbose
        self._tmp_dir = tmp_dir
        self._algo_name = algo
        self._algo_class = ALGORITHM_MAP[algo]

        # Dynamically instantiate algorithm
        self.algo = getattr(sys.modules[__name__], self._algo_class)(
            data=data,
            metric=target_metric,
            cv_n_jobs=CV_N_JOBS,
            hpo_algo=hpo_algo,
            verbose=verbose,
        )

    def start(self, max_evals):

        try:
            with ResultLoggingManager(self._tmp_dir, self._algo_class) as result_logger:
                self.algo.set_result_logger(result_logger)

                best, trials = self.algo.best_hyperparams(max_evals)

                best_trial = None
                # Sort best trial based on loss value
                if best:
                    best_trial = sorted(
                        trials.results, key=lambda x: x["loss"], reverse=False
                    )[0]
                else:
                    best_trial = trials

                return best, best_trial
        except Exception as e:
            self.__logger.ERROR("Exception : ", e)
            self.__logger.ERROR(traceback.format_exc())

            return False, False

    def start_with_limits(self, max_evals, time_limit, tasks):
        """
        Start the trainer with a time limit
        """
        with ResultLoggingManager(self._tmp_dir, self._algo_class) as result_logger:
            try:
                # Although hyperopt.fmin does have a timout parameter, it doesnt seem to work with multiprocessing
                with limits.run_with_enforced_limits(time_limit=time_limit):
                    self.algo.set_result_logger(result_logger)

                    _, best_trial = self.algo.best_hyperparams(max_evals)

                    if self.algo.trials.results:
                        best_trial = sorted(
                            self.algo.trials.results,
                            key=lambda x: x["loss"],
                            reverse=False,
                        )[0]

                    tasks[self._algo_name] = {
                        **best_trial,
                        "exception": False
                    }

            except TimeoutException:
                # Handle timeout when enforced cpu time limit is reached
                trials = self.algo.trials

                if trials.results:
                    best_trial = sorted(
                        trials.results, key=lambda x: x["loss"], reverse=False
                    )[0]
                    # A timeout exception is not considered an algorithm exception
                    tasks[self._algo_name] = {
                        **best_trial,
                        "exception": False
                    }
                else:
                    # When no trials were completed before the job timed out
                    tasks[self._algo_name] = {
                        "loss": None, 
                        "hyperparams": None,
                        "exception": False,
                    }

            except Exception:
                self.__logger.error("Exception for algo {0}".format(self._algo_name))
                self.__logger.error(traceback.format_exc())

                tasks[self._algo_name] = {
                    "loss": None, 
                    "hyperparams": None,
                    "exception": True,
                }



from hyperopt import fmin, tpe, hp, STATUS_OK, Trials
from surprise import NormalPredictor
from surprise.model_selection import cross_validate
from auto_surprise.constants import DEFAULT_MAX_EVALS, ACCURACY_METRICS
from auto_surprise.algorithms.base import AlgorithmBase


class AutoSurpriseNormalPredictor(AlgorithmBase):
    """
    Wrapper for surprise.prediction_algorithms.random_pred.NormalPredictor

    This model is used as a baseline to eliminate models that perform worse
    than random.
    Does not require any hyperparameter tuning
    """

    def _hyperopt(self):
        algo = NormalPredictor()
        return cross_validate(
            algo,
            self._data,
            measures=ACCURACY_METRICS,
            cv=self._cv,
            n_jobs=self._cv_n_jobs,
            verbose=self.verbose,
        )[self._metric].mean()

    def _objective(self):
        loss = self._hyperopt()
        self._result_logger.append_results(loss, None)

        return {"loss": loss, "status": STATUS_OK, "hyperparams": None}

    def best_hyperparams(self, max_evals=DEFAULT_MAX_EVALS):
        best = self._objective()
        return None, best



import unittest
import utils as test_utils
from auto_surprise.engine import Engine
from auto_surprise.constants import FULL_ALGO_LIST

class TestEngine(unittest.TestCase):
    
    def setUp(self):
        self.data = test_utils.load_test_dataset()
        self.engine = Engine()

    def test_trainer(self):

        with self.subTest("Sanity - everything just works"):
            best_algo, best_params, best_score, tasks=self.engine.train(data=self.data, cpu_time_limit=60, max_evals=20)
            self.assertTrue(best_algo)
            self.assertTrue(best_score)
            self.assertTrue(tasks)

            self.assertCountEqual(list(tasks.keys()), FULL_ALGO_LIST)



import setuptools

install_requires = [
    "hyperopt",
    "lightgbm",
    "numpy",
    "scikit-learn==0.22.0",
    "scikit-surprise"
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="auto-surprise",
    version="0.1.4",
    author="Rohan Anand",
    author_email="anandr@tcd.ie",
    description="A python package that automates algorithm selection and hyperparameter tuning for the recommender system library Surprise",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BeelGroup/Auto-Surprise",
    download_url="https://github.com/BeelGroup/Auto-Surprise/archive/v0.1.4.tar.gz",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=install_requires,
)



from auto_surprise.constants import (
    DEFAULT_TARGET_METRIC,
    CV_N_JOBS,
    DEFAULT_HPO_ALGO,
    DEFAULT_CV_ITERS,
)
from hyperopt import Trials


class AlgorithmBase(object):
    def __init__(
        self,
        cv=DEFAULT_CV_ITERS,
        metric=DEFAULT_TARGET_METRIC,
        data=None,
        cv_n_jobs=CV_N_JOBS,
        hpo_algo=DEFAULT_HPO_ALGO,
        verbose=False,
    ):
        self._cv = cv
        self._metric = metric
        self._data = data
        self.verbose = verbose
        self._cv_n_jobs = cv_n_jobs
        self._hpo_algo = hpo_algo
        self.trials = Trials()

    def set_result_logger(self, result_logger_manager):
        self._result_logger = result_logger_manager
